from pycoingecko import CoinGeckoAPI
from requests.exceptions import HTTPError

from .. import AlgoLabAPI
from ...tradables import AlgoLabTradable, AlgoLabCryptoTradable


TRADABLE_TABLE = {
    AlgoLabCryptoTradable.BTC: "bitcoin",
    AlgoLabCryptoTradable.BSV: "bitcoin-cash-sv",
    AlgoLabCryptoTradable.LTC: "litecoin",
    AlgoLabCryptoTradable.ETH: "ethereum",
}


class AlgoLabCoinGeckoAPI(AlgoLabAPI):
    def __init__(self):
        self._api = CoinGeckoAPI()

    def up(self) -> bool:
        """
        returns true or false if service is up
        """
        try:
            self._api.ping()['gecko_says']
            return True
        except (KeyError, HTTPError):
            return False

    def get_price(self, tradable: AlgoLabCryptoTradable, currency: str = 'usd') -> tuple:
        """
        get the current price of tradable in `currency`
        returns a tuple of (last_updated, current_price)
        """
        coin_data = self._api.get_coin_by_id(id=TRADABLE_TABLE[tradable])
        last_updated = coin_data['last_updated']
        current_price = coin_data['market_data']['current_price'][currency]
        return (last_updated, current_price)



from threading import Thread
import time
from uuid import uuid4, UUID

from .listener import AlgoLabPriceListener


class AlgoLabPriceListenerService():
    def __init__(self):
        self._listeners: {UUID} = {}
        self._thread = Thread(target=self._thread_loop)
        self._running = False

    def _start(self):
        """
        start a new thread for listening
        """
        self._running = True
        self._thread.start()

    def _stop(self, timeout=None):
        """
        stop the listening thread
        """
        self._running = False
        self._thread.join(timeout=timeout)

    def _thread_loop(self):
        while self._running:
            for key in list(self._listeners):
                self._listeners[key]._update()
                time.sleep(1)

    def register(self, listener: AlgoLabPriceListener) -> UUID:
        """
        returns UUID of item to be able to deregister later
        """
        uuid = uuid4()
        self._listeners[uuid] = listener
        # The listener manages its own API, so that a
        # user may extend the behavior, we just
        # manage the update frequency for resource
        # util management
        listener._update()
        return uuid

    def deregister(self, uuid: UUID):
        """
        deregister listener from service

        args:
            uuid: the uuid returned when registering the listener
        """
        del self._listeners[uuid]



from ...price_listener.service import AlgoLabPriceListenerService


class AlgoLabConsumerAPI():
    def __init__(self, listener_service: AlgoLabPriceListenerService):
        """
        docstring
        """
        self._listener_service = listener_service

    def get_listener_service(self) -> AlgoLabPriceListenerService:
        return self._listener_service



from ..apis import AlgoLabAPI
from ..tradables import AlgoLabTradable


class AlgoLabPriceListener():
    def __init__(self, api: AlgoLabAPI, tradable: AlgoLabTradable, callback: callable, update_frequency=None):
        self._current_data = None
        self._last_updated = None
        self._history = []

        self._update_frequency = update_frequency
        self._api = api
        self._tradable = tradable
        self._callback = callback

    def _change(self):
        """
        passes callback the data.
        data is (timestamp, price)
        """
        self._callback(self._current_data)

    def _update(self):
        """
        The listener manages the update behavior for
        user extensibility. Timing the call of this event
        is the scope of the service
        """
        last_updated, data = self._api.get_price(self._tradable)
        if not self._last_updated \
                or not self._current_data \
                or self._last_updated > last_updated:
            self._last_updated = last_updated
            self._current_data = (last_updated, data)
            self._change()



from threading import Thread

from .abc_algo import AlgoLabAlgorithm
from .apis.consumer import AlgoLabConsumerAPI
from .price_listener.service import AlgoLabPriceListenerService


class AlgoLabService():
    def __init__(self):
        """
        The whole service
        It runs shit
        """
        self._listener_service = AlgoLabPriceListenerService()
        self._consumer_api = AlgoLabConsumerAPI(self._listener_service)
        self._running = False
        self._running_algos = []
        self._thread = Thread(target=self._thread_loop)

    def _start(self):
        """
        start a new thread for work
        """
        self._running = True
        self._listener_service._start()
        self._thread.start()

    def _stop(self, timeout=None):
        """
        stop the listening thread
        """
        for algo in self._running_algos:
            algo.__destroy__()

        self._running = False
        self._listener_service._stop(timeout=timeout)
        self._thread.join(timeout=timeout)

    def _thread_loop(self):
        while self._running:
            pass

    def _run_algo(self, algo: AlgoLabAlgorithm):
        self._running_algos.append(algo(self._consumer_api))



from abc import ABC, abstractclassmethod
from ..tradables import AlgoLabTradable


class AlgoLabAPI(ABC):
    @abstractclassmethod
    def get_price(self, tradable: AlgoLabTradable, currency: str = 'usd'):
        """
        get the current price in `currency`
        """
        pass

    @abstractclassmethod
    def up(self) -> bool:
        """
        returns true or false if service is up
        """
        pass



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("version", "r") as fh:
    version = fh.read()

setuptools.setup(
    name="algolab",
    version=version,
    author="Jacob McSwain",
    author_email="pypi@mcswain.dev",
    description="AlgoLab Site API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/USA-RedDragon/python-algolab",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
    ],
    python_requires='>=3.6',
)
